/** @file puzzle.c
 * Implementation of the <b>Puzzle</b> class.
 */
#include "puzzle.h"
//#include <string.h>
#include <assert.h>
#include <stdio.h> // printf()
#include <ctype.h> // isdigit()

/** 
 * @brief Assign a value to a cell.
 *
 * @param p the puzzle
 * @param r row
 * @param c column
 * @param v value to assign
 * @return 1 if the assignment succeeded<br/>
 * 	0 otherwise
 *
 * It automatically removes possible candidates from other cells in the same unit.
 *
 * @see _Puzzle::g
 */
int puzzle_set_cell(Puzzle *p, int r, int c, int v)
{
	assert(p);
	assert(0 <= r && r < 9 && 0 <= c && c < 9 && 1 <= v && v <= 9);

	// Check if the cell isn't already set or if @p v is a candidate.
	if(p->g[r][c] || !p->sg[r][c][v])
		return 0;

	int i, j, rr = r/3*3, rc = c/3*3;
	for(i = 0; i < 9; ++i) { // Apply the One Rule to ...
		p->sg[r][i][v] = 0; // row,
		p->sg[i][c][v] = 0; // column,
		p->sg[r][c][i+1] = 0; // the cell at [r][c],
	}
	for(i = 0; i < 3; ++i) for(j = 0; j < 3; ++j)
		p->sg[rr+i][rc+j][v] = 0; // region.

	p->g[r][c] = v;
	p->sg[r][c][v] = 1;
	p->progress++;

	return 1;
}

/** 
 * @brief Remove a value as possible candidate for a cell.
 *
 * @param p the puzzle
 * @param r row
 * @param c column
 * @param v value to remove
 * @return 1 if the operation succeeded<br/>
 * 	0 otherwise
 *
 * @see _Puzzle::g
 */
int puzzle_rm_candidate(Puzzle *p, int r, int c, int v)
{
	assert(p);
	assert(0 <= r && r < 9 && 0 <= c && c < 9 && 1 <= v && v <= 9);

	// Check if the cell isn't already set or if @p v is a candidate.
	if(p->g[r][c] || !p->sg[r][c][v])
		return 0;
	p->sg[r][c][v] = 0;
	return 1;
}

/** 
 * @brief Assign a whole grid to a puzzle.
 *
 * @param p the puzzle
 * @param g the grid to assign
 * @return 1 on success (the grid represents a puzzle respecting the One Rule)<br>
 * 	0 otherwise
 *
 * @warning @p g must be contiguous
 * 	- see http://www.c-faq.com/aryptr/ary2dfunc3.html
 * 	- see http://www.c-faq.com/aryptr/dynmuldimary.html
 */
int puzzle_set_grid(Puzzle *p, const int *g)
{
	assert(p && g);
	puzzle_clear(p);
	int i, j;
	for(i = 0; i < 9; ++i) for(j = 0; j < 9; ++j) if(g[i*9+j]) {
		//printf("%d %d %d\n", i, j, g[i*9+j]);
		if(!puzzle_set_cell(p, i, j, g[i*9+j]))
			return 0;
	}
	return 1;
}

/** 
 * @brief Assign a whole string-formatted grid to a puzzle.
 *
 * @param p the puzzle
 * @param s the string representing the grid to assign
 * @return 1 on success (the string represents a puzzle respecting the One Rule)<br>
 * 	0 otherwise
 *
 * @warning @p s must have at least 81 characters.
 * @see puzzle_str2grid()
 */
int puzzle_set_grid_str(Puzzle *p, const char *s)
{
	assert(p && s);
	int tmp[9][9];
	puzzle_str2grid(s, *tmp);
	return puzzle_set_grid(p, *tmp);
}

/**
 * @brief Clear puzzle.
 *
 * @param p the puzzle to clear
 *
 * @warning Always clear a ::Puzzle before using it!
 */
void puzzle_clear(Puzzle *p)
{
	assert(p);

	int i, j, k;
	for(i = 0; i < 9; ++i) for(j = 0; j < 9; ++j) {
		p->g[i][j] = 0;
		for(k = 1; k <= 9; ++k)
			p->sg[i][j][k] = 1;
	}
	p->progress = 0;
}

/** 
 * @brief Check if the puzzle is valid.
 *
 * @param p the puzzle to check
 * @return 1 if the puzzle is valid<br>
 * 	0 otherwise
 *
 * By construction, a Puzzle object never has a digit more than once in each row, column or region. However, the puzzle may not be solvable (there is a dead end), hence it is not valid.
 */
int puzzle_valid(const Puzzle *p)
{
	assert(p);

	int i, j, k, cnt;
	for(i = 0; i < 9; ++i) for(j = 0; j < 9; ++j) {
		for(cnt = 0, k = 1; k <= 9; ++k)
			cnt += p->sg[i][j][k];
		if(cnt == 0)
			return 0;
	}
	return 1;
}

/** 
 * @brief Check if the puzzle is solved.
 *
 * @param p the puzzle
 * @return 1 if the puzzle is complete<br/>
 * 	0 otherwise
 */
int puzzle_solved(const Puzzle *p)
{
	assert(p);
	return p->progress == 81;
}

/** 
 * @brief Print the grid nicely to stdout.
 *
 * @param p the puzzle to print
 */
void puzzle_print(const Puzzle *p)
{
	assert(p);

	int i, j;
	for(i = 0; i < 9; ++i) {
		for(j = 0; j < 9; ++j) {
			printf("%c ", p->g[i][j]? p->g[i][j]+'0' : ' ');
			if(j == 2 || j == 5)
				printf("|");
		}
		if(i == 2 || i == 5)
			printf("\n------+------+------");
		printf("\n");
	}
}

/** 
 * @brief Convert a puzzle from string representation to grid.
 *
 * @param s the source string
 * @param g the destination grid
 *
 * A string representation of a puzzle consists of 81 digits from 1 to 9 or any other character to denote an empty cell.
 * @warning @p s must have at least 81 characters.
 * @warning The method doesn't do any checking, it simply converts from one format to another.
 * @warning @p g must be contiguous
 * 	- see http://www.c-faq.com/aryptr/ary2dfunc3.html
 * 	- see http://www.c-faq.com/aryptr/dynmuldimary.html
 */
void puzzle_str2grid(const char *s, int *g)
{
	assert(s && g);
	for(int i = 0; i < 81; ++i)
		*(g+i) = isdigit(s[i])? s[i]-'0' : 0;
}
