/** @file puzzle.h
 * Interface of the <b>Puzzle</b> class.
 */

#ifndef _PUZZLE_H
#define _PUZZLE_H

/**
 * @addtogroup Puzzle_Group Puzzle
 *
 * Class that simulates a Sudoku puzzle.
 *
 * @section Puzzle_Sec_Spec Specification
 *
 * <p>
 * A Sudoku is a <b>grid</b> of 9x9 <b>cells</b> completed with digits from 1 to 9 such that each digit appears exactly once in each <b>row</b>, <b>column</b> or <b>region</b>.<br/>
 * A region is one of the nine 3x3 sub-grids.<br/>
 * Rows, columns and regions are collectively referred to as <b>units</b>.<br/>
 * A Sudoku <b>puzzle</b> consists of a partially completed grid.<br/>
 * The <i>One Rule</i> can then be compactly stated as: 'Each digit appears once in each unit'.
 * </p>
 *
 * The Puzzle class offers a way to store and manipulate a Sudoku puzzle. It guarantees that by using its public interface the puzzle will satisfy at any given time the constraint that a digit appears at _most_ one time in each row, column or region. However, this does not insure that the puzzle is valid (eg: it has a solution).
 *
 * @section Puzzle_Sec_Man Manual
 *
 * Example 1:
 * Basics
 * @code
 * Puzzle p;
 * puzzle_clear(&p); // Always clear a Puzzle before using it!
 * puzzle_set_cell(&p, 0, 0, 3); // OK; notice that cell 0,0 is in the upper-left corner
 * puzzle_set_cell(&p, 0, 8, 3); // error: there is already a 3 on the same row
 * puzzle_set_cell(&p, 5, 0, 3); // error: there is already a 3 on the same column
 * puzzle_set_cell(&p, 1, 1, 3); // error: there is already a 3 in the same region
 * puzzle_set_cell(&p, 0, 0, 1); // error: cell (0,0) is already set
 * puzzle_set_cell(&p, 0, 1, 2); // OK
 * puzzle_print(&p);
 * @endcode
 *
 * Example 2:
 * How to check if a given puzzle respects the One Rule
 * @code
 * const int s[9][9] = {
 * 	{8,6,1,0,0,0,3,0,0},
 * 	{0,0,0,5,6,3,0,0,0},
 * 	{0,0,0,0,9,0,2,7,6},
 * 	{7,2,0,8,0,1,0,0,0},
 * 	{4,9,5,0,0,0,8,1,3},
 * 	{0,0,0,9,0,4,0,6,2},
 * 	{3,8,4,0,1,0,0,0,0},
 * 	{0,0,0,2,8,9,0,0,0},
 * 	{0,0,9,0,0,0,6,8,5}
 * };
 * Puzzle p;
 * if(puzzle_set_grid(&p, *s)) { // puzzle_set_grid() takes care of clearing the puzzle
 * 	printf("Puzzle is valid.\n");
 * 	puzzle_print(&p);
 * } else printf("Puzzle is not valid.\n");
 * @endcode
 *
 * Example 3:
 * The same but using a string
 * @code
 * char s[] = "4.....8.5.3..........7......2.....6.....8.4......1.......6.3.7.5..2.....1.4......";
 * Puzzle p;
 * if(puzzle_set_grid_str(&p, s)) { // puzzle_set_grid_str() takes care of clearing the puzzle
 * 	printf("Puzzle is valid.\n");
 * 	puzzle_print(&p);
 * } else printf("Puzzle is not valid.\n");
 * @endcode
 *
 * @{
 */

/**
 * @typedef Puzzle
 * Container which stores a Sudoku puzzle.
 */
typedef struct _Puzzle Puzzle;

int puzzle_set_cell(Puzzle *p, int r, int c, int v);
int puzzle_rm_candidate(Puzzle *p, int r, int c, int v);
int puzzle_set_grid(Puzzle *p, const int *g);
int puzzle_set_grid_str(Puzzle *p, const char *s);
void puzzle_clear(Puzzle *p);

static inline int puzzle_get_cell(const Puzzle *p, int r, int c);
static inline int puzzle_get_candidate(const Puzzle *p, int r, int c, int v);
int puzzle_valid(const Puzzle *p);
int puzzle_solved(const Puzzle *p);

void puzzle_print(const Puzzle *p);
void puzzle_str2grid(const char *s, int *g);

/**
 * @struct _Puzzle
 */
struct _Puzzle {
	/**
	 * @warning The rows and columns are 0-indexed.<br/>
	 * The upper-left corner is (0, 0).<br/>
	 * The lower-right corner is (8, 8).<br/>
	 */
	int g[9][9]; ///< Grid
	int sg[9][9][10]; ///< Cell-state grid
	int progress; ///< Progress
};

#include "puzzle.x"

/**
 * @}
 */

#endif /* _PUZZLE_H */
