/**
 * @file solver.c
 * Implementation of <b>Solver</b> module.
 */

#include "puzzle.h"
#include "solver.h"
#include <stdio.h> // printf
#include <string.h>// strlen

#define GET(p, i, j) (puzzle_get_cell((p), (i), (j)))
#define GETC(p, i, j, k) (puzzle_get_candidate((p), (i), (j), (k)))
#define SET(p, i, j, k) (puzzle_set_cell((p), (i), (j), (k)))
#define RMC(p, i, j, k) (puzzle_rm_candidate((p), (i), (j), (k)))

static char space[100];
/** 
 * @brief Solve any (valid) puzzle.
 *
 * @param p the puzzle to solve
 * @return 1 if a solution has been found<br/>
 * 	0 otherwise
 *
 * The method attempts to solve the puzzle using basic logic and falls back to brute force when no more deductions can be made.<br/>
 * @p p will be filled as far as the algorithm can go. If there is at least one solution, the algorithm should be able to find it.
 */
int solver_solve(Puzzle *p)
{
  space[0]='\0';
  // Use logic
	while(_naked_singles(p) || _hidden_singles(p));
	if(puzzle_solved(p))
		return 1;
	// May the (brute) force be with you!
	Puzzle copy;
	int g[3]; // guess
	while(_make_guess(p, g)) {
	  printf("%sMaking a guess %d %d %d\n",space, g[0], g[1], g[2]);
		copy = *p;
		strcat(space," "); //space ++
		SET(&copy, g[0], g[1], g[2]);
		if(solver_solve(&copy)) {
			*p = copy;
			return 1;
		}
		else {
		  space[strlen(space)]='\0'; //space --
		  RMC(p, g[0], g[1], g[2]);
		}
	}
	return 0;
}

/** 
 * @brief <i>Internal function</i>.
 * Solve a puzzle by searching for naked singles.
 *
 * @param p the puzzle to solve
 * @return the number of cells this algorithm managed to assign
 *
 * A naked single is the single candidate in a cell.
 */
int _naked_singles(Puzzle *p)
{
	int i, j, k, oldk = 0, cnt, n = 0;
	for(i = 0; i < 9; ++i) for(j = 0; j < 9; ++j) if(!GET(p, i, j)) {
		for(cnt = 0, k = 1; k <= 9; ++k) if(GETC(p, i, j, k))
			++cnt, oldk = k;
		if(cnt == 1) // found a naked single
			++n, SET(p, i, j, oldk);
	}
	return n;
}

/** 
 * @brief <i>Internal function</i>.
 * Solve a puzzle by searching for hidden singles.
 *
 * @param p the puzzle to solve
 * @return the number of cells this algorithm managed to assign
 *
 * A hidden single is a candidate that appears with others, but only once in a given row, column or box.
 */
int _hidden_singles(Puzzle *p)
{
	int i, j, k, oldi = 0, oldj = 0, ri, rj, cnt, n = 0;
	for(k = 1; k <= 9; ++k) {
		// Search rows
		for(i = 0; i < 9; ++i) {
			for(cnt = j = 0; j < 9; ++j) if(!GET(p, i, j) && GETC(p, i, j, k))
				++cnt, oldj = j;
			if(cnt == 1) // found a hidden single
				++n, SET(p, i, oldj, k);
		}
		// Search columns
		for(j = 0; j < 9; ++j) {
			for(cnt = i = 0; i < 9; ++i) if(!GET(p, i, j) && GETC(p, i, j, k))
				++cnt, oldi = i;
			if(cnt == 1) // found a hidden single
				++n, SET(p, oldi, j, k);
		}
		// Search regions
		for(ri = 3; ri <= 9; ri += 3) for(rj = 3; rj <= 9; rj += 3) {
			for(cnt = 0, i = ri-3; i < ri; ++i) for(j = rj-3; j < rj; ++j) if(!GET(p, i, j) && GETC(p, i, j, k))
				++cnt, oldi = i, oldj = j;
			if(cnt == 1) // found a hidden single
				++n, SET(p, oldi, oldj, k);
		}
	}
	return n;
}

/** 
 * @brief <i>Internal function</i>.
 * Attempt to guess the value of a cell.
 *
 * @param p the puzzle
 * @param guess the best guess found (row, column, value) if there is any choice left
 * @return 1 if a there is at least one choice
 * 	0 if there is no solution
 */
int _make_guess(const Puzzle *p, int guess[3])
{
	int i, j, k, oldk = 0, cnt, bstcnt = 10;
	for(i = 0; i < 9; ++i) for(j = 0; j < 9; ++j) if(!GET(p, i, j)) {
		for(cnt = 0, k = 1; k <= 9; ++k) if(GETC(p, i, j, k))
			++cnt, oldk = k;
		if(cnt < bstcnt)
			bstcnt = cnt, guess[0] = i, guess[1] = j, guess[2] = oldk;
	}
	return bstcnt != 0;
}

#undef GET
#undef GETC
#undef SET
#undef RMC
